from __future__ import annotations

from typing import TYPE_CHECKING

from sila2.framework.abc.constraint import Constraint

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class MaximalElementCount(Constraint):
    max_elements: int

    def __init__(self, max_elements: int):
        self.max_elements = max_elements

    def validate(self, value: list) -> bool:
        return len(value) <= self.max_elements

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> MaximalElementCount:
        return cls(int(fdl_node.text))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.max_elements})"



from os.path import dirname, join

from sila2.framework import Feature

AuthorizationProviderServiceFeature = Feature(
    open(join(dirname(__file__), "AuthorizationProviderService.sila.xml")).read()
)



from __future__ import annotations

from typing import NamedTuple


class Verify_Responses(NamedTuple):

    TokenLifetime: int
    """
    The lifetime of the provided access token as the maximum validity period after the last SiLA Server request.
    """



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from sila2.framework.abc.constraint import Constraint

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class Pattern(Constraint):
    pattern: str

    def __init__(self, pattern: str):
        self.pattern = pattern

    def validate(self, value: str) -> bool:
        return bool(re.fullmatch(self.pattern, value))

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> Pattern:
        return cls(fdl_node.text)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.pattern!r})"



from __future__ import annotations

import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Dict, List, Optional, Union
from uuid import UUID

import grpc

from sila2.discovery.broadcaster import SilaServiceBroadcaster
from sila2.discovery.service_info import SilaServiceInfo
from sila2.framework.binary_transfer.server_binary_transfer_handler import ServerBinaryTransferHandler
from sila2.framework.constraints.maximal_length import MaximalLength
from sila2.framework.constraints.pattern import Pattern
from sila2.framework.feature import Feature
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.utils import parse_feature_definition, xpath_sila
from sila2.server.encryption import generate_self_signed_certificate
from sila2.server.feature_implementation_base import FeatureImplementationBase
from sila2.server.feature_implementation_servicer import FeatureImplementationServicer
from sila2.server.metadata_interceptor import MetadataInterceptor

if TYPE_CHECKING:
    from sila2.framework.utils import HasFullyQualifiedIdentifier


class SilaServer:
    grpc_server: grpc.Server
    features: Dict[str, Feature]
    feature_servicers: Dict[str, FeatureImplementationServicer]
    running: bool
    generated_ca: Optional[bytes]

    server_name: str
    server_type: str
    server_uuid: UUID
    server_description: str
    server_version: str
    server_vendor_url: str

    service_broadcaster: SilaServiceBroadcaster
    running_instances: List[SilaServiceInfo]

    binary_transfer_handler: ServerBinaryTransferHandler

    metadata_interceptors: List[MetadataInterceptor]

    children_by_fully_qualified_identifier: Dict[FullyQualifiedIdentifier, HasFullyQualifiedIdentifier]

    def __init__(
        self,
        server_name: str,
        server_type: str,
        server_description: str,
        server_version: str,
        server_vendor_url: str,
        server_uuid: Optional[Union[str, UUID]] = None,
        max_grpc_workers: int = 100,
        max_child_task_workers: int = 100,
    ):
        self.running = False
        self.logger = logging.getLogger(__name__)
        self.generated_ca = None

        # import locally to prevent circular import
        from sila2.features.silaservice import SiLAServiceFeature
        from sila2.server.silaservice_impl import SiLAServiceImpl

        silaservice_fdl_tree = parse_feature_definition(SiLAServiceFeature._feature_definition)

        name_constraint = MaximalLength(
            int(
                xpath_sila(
                    silaservice_fdl_tree,
                    "//sila:Command[sila:Identifier/text() = 'SetServerName']//sila:MaximalLength/text()",
                )[0]
            )
        )
        type_constraint = Pattern(
            xpath_sila(
                silaservice_fdl_tree, "//sila:Property[sila:Identifier/text() = 'ServerType']//sila:Pattern/text()"
            )[0]
        )
        version_constraint = Pattern(
            xpath_sila(
                silaservice_fdl_tree, "//sila:Property[sila:Identifier/text() = 'ServerVersion']//sila:Pattern/text()"
            )[0]
        )
        vendor_url_constraint = Pattern(
            xpath_sila(
                silaservice_fdl_tree, "//sila:Property[sila:Identifier/text() = 'ServerVendorURL']//sila:Pattern/text()"
            )[0]
        )

        if not name_constraint.validate(server_name):
            raise ValueError(f"Server name {server_name!r} does not satisfy constraint {name_constraint!r}")
        if not type_constraint.validate(server_type):
            raise ValueError(f"Server type {server_type!r} does not satisfy constraint {type_constraint!r}")
        if not version_constraint.validate(server_version):
            raise ValueError(
                f"Server version {server_version!r} does not satisfy constraint {version_constraint!r}. "
                f"Examples: '2.1', '0.1.3', '1.2.3_preview'"
            )
        if not vendor_url_constraint.validate(server_vendor_url):
            raise ValueError(
                f"Server vendor url {server_vendor_url!r} does not satisfy constraint {vendor_url_constraint!r}"
            )

        self.server_name = server_name
        self.server_type = server_type
        if server_uuid is None:
            self.server_uuid = uuid.uuid4()
        else:
            self.server_uuid = server_uuid if isinstance(server_uuid, UUID) else UUID(server_uuid)
        self.server_description = server_description
        self.server_version = server_version
        self.server_vendor_url = server_vendor_url

        self.grpc_server = grpc.server(
            ThreadPoolExecutor(max_workers=max_grpc_workers, thread_name_prefix=f"grpc-executor-{self.server_uuid}")
        )
        self.service_broadcaster = SilaServiceBroadcaster()
        self.running_instances = []
        self.features = {}
        self.feature_servicers = {}
        self.metadata_interceptors = []

        self.binary_transfer_handler = ServerBinaryTransferHandler(self.grpc_server)

        self.child_task_executor = ThreadPoolExecutor(
            max_workers=max_child_task_workers, thread_name_prefix=f"child-task-executor-{self.server_uuid}"
        )

        self.children_by_fully_qualified_identifier = {}

        self.set_feature_implementation(SiLAServiceFeature, SiLAServiceImpl(parent_server=self))

    def set_feature_implementation(self, feature: Feature, implementation: FeatureImplementationBase):
        if feature._identifier in self.feature_servicers:
            self.feature_servicers[feature._identifier].implementation.stop()

        class FeatureServicer(FeatureImplementationServicer, feature._servicer_cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._set_implementation(implementation)

        servicer: FeatureImplementationServicer = FeatureServicer(self, feature)
        self.features[feature._identifier] = feature
        self.feature_servicers[feature._identifier] = servicer

        # add self as servicer to the gRPC server
        getattr(feature._grpc_module, f"add_{feature._identifier}Servicer_to_server")(servicer, self.grpc_server)

        feature._binary_transfer_handler = self.binary_transfer_handler
        self.children_by_fully_qualified_identifier[feature.fully_qualified_identifier] = feature
        self.children_by_fully_qualified_identifier.update(feature._children_by_fully_qualified_identifier)

    def add_metadata_interceptor(self, interceptor: MetadataInterceptor):
        self.metadata_interceptors.append(interceptor)

    def __start(
        self,
        address: str,
        port: int,
        *,
        insecure: bool,
        private_key: Optional[bytes] = None,
        cert_chain: Optional[bytes] = None,
        enable_discovery: bool = True,
    ):
        # start implementations
        for servicer in self.feature_servicers.values():
            servicer.implementation.start()

        # start grpc server
        address_string = f"{address}:{port}"
        self.generated_ca = None
        if insecure:
            self.logger.warning("Starting SiLA server without encryption. This violates the SiLA 2 specification.")
            self.grpc_server.add_insecure_port(address_string)
        else:
            if private_key is None and cert_chain is None:
                if not enable_discovery:
                    raise ValueError("If discovery is disabled, private key and certificate chain are required")

                self.logger.info("Generating self-signed certificate")
                private_key, cert_chain = generate_self_signed_certificate(self.server_uuid, address)
                self.generated_ca = cert_chain
            if private_key is None or cert_chain is None:
                raise ValueError(
                    "For secure connections, either provide both the private key and certificate chain, "
                    "or none of them (server will then generate a self-signed certificate)"
                )
            self.logger.info("Starting SiLA server with encryption")
            credentials = grpc.ssl_server_credentials([(private_key, cert_chain)])
            self.grpc_server.add_secure_port(address_string, server_credentials=credentials)
        self.grpc_server.start()

        # start zeroconf broadcasting
        if enable_discovery:
            self.logger.info("Starting zeroconf broadcasting for SiLA Server Discovery")
            info = self.service_broadcaster.register_server(self, address, port, ca=self.generated_ca)
            self.running_instances.append(info)

        self.running = True

    def start_insecure(self, address: str, port: int, enable_discovery: bool = True):
        self.__start(address, port, insecure=True, enable_discovery=enable_discovery)

    def start(
        self,
        address: str,
        port: int,
        private_key: bytes = None,
        cert_chain: bytes = None,
        enable_discovery: bool = True,
    ):
        self.__start(
            address,
            port,
            insecure=False,
            private_key=private_key,
            cert_chain=cert_chain,
            enable_discovery=enable_discovery,
        )

    def stop(self, grace_period: Optional[float] = None):
        # stop zeroconf broadcasting
        for info in self.running_instances:
            self.service_broadcaster.unregister_server(info)
            self.running_instances.remove(info)

        # stop feature implementations
        for servicer in self.feature_servicers.values():
            servicer.cancel_all_subscriptions()
            servicer.implementation.stop()
        self.child_task_executor.shutdown(wait=True)

        # stop grpc server
        self.grpc_server.stop(grace_period).wait()

        self.running = False

    def __getitem__(self, item: str) -> Feature:
        return self.features[item]

    def __del__(self):
        try:
            if self.running:
                self.stop()
        except AttributeError:  # when __init__ fails, some attributes might not exist yet, then they cannot be stopped
            pass



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.client_metadata import ClientMetadataInstance
from sila2.client.utils import call_rpc_function
from sila2.framework.command.unobservable_command import UnobservableCommand

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature

T = TypeVar("T")


class ClientUnobservableCommand(Generic[T]):
    _parent_feature: ClientFeature
    _wrapped_command: UnobservableCommand

    def __init__(self, parent_feature: ClientFeature, wrapped_command: UnobservableCommand):
        self._parent_feature = parent_feature
        self._wrapped_command = wrapped_command

    def __call__(self, *args, **kwargs) -> T:
        raw_metadata: Optional[Iterable[ClientMetadataInstance]] = kwargs.pop("metadata", None)
        param_msg = self._wrapped_command.parameters.to_message(
            *args, **kwargs, toplevel_named_data_node=self._wrapped_command.parameters
        )
        response_msg = call_rpc_function(
            getattr(self._parent_feature._grpc_stub, self._wrapped_command._identifier),
            param_msg,
            metadata=raw_metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_command,
        )

        return self._wrapped_command.responses.to_native_type(response_msg)



from __future__ import annotations

from types import ModuleType
from typing import TYPE_CHECKING, Dict, Optional, Type, Union

from sila2.framework.abc.named_node import NamedNode
from sila2.framework.command.command import Command
from sila2.framework.command.observable_command import ObservableCommand
from sila2.framework.command.unobservable_command import UnobservableCommand
from sila2.framework.data_types.data_type_definition import DataTypeDefinition
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.metadata import Metadata
from sila2.framework.property.observable_property import ObservableProperty
from sila2.framework.property.property import Property
from sila2.framework.property.unobservable_property import UnobservableProperty
from sila2.framework.utils import (
    feature_definition_to_modules,
    parse_feature_definition,
    xml_node_to_normalized_string,
    xpath_sila,
)

if TYPE_CHECKING:
    from sila2.framework.abc.binary_transfer_handler import BinaryTransferHandler
    from sila2.framework.utils import HasFullyQualifiedIdentifier


class Feature(NamedNode):
    _feature_definition: str
    fully_qualified_identifier: FullyQualifiedIdentifier
    _sila2_version: str
    _feature_version: str
    _maturity_level: str
    _locale: str
    _originator: str
    _category: str
    _pb2_module: ModuleType
    _grpc_module: ModuleType
    _servicer_cls: Type
    _observable_properties: Dict[str, ObservableProperty]
    _unobservable_properties: Dict[str, UnobservableProperty]
    _observable_commands: Dict[str, ObservableCommand]
    _unobservable_commands: Dict[str, UnobservableCommand]
    _data_type_definitions: Dict[str, DataTypeDefinition]
    defined_execution_errors: Dict[str, DefinedExecutionErrorNode]
    metadata_definitions: Dict[str, Metadata]
    _binary_transfer_handler: Optional[BinaryTransferHandler]
    _children_by_fully_qualified_identifier: Dict[FullyQualifiedIdentifier, HasFullyQualifiedIdentifier]

    def __init__(self, feature_definition: str) -> None:
        fdl_node = parse_feature_definition(feature_definition)
        super().__init__(fdl_node)

        self._feature_definition = xml_node_to_normalized_string(fdl_node)

        self._sila2_version = fdl_node.attrib["SiLA2Version"]
        self._feature_version = fdl_node.attrib["FeatureVersion"]
        self._originator = fdl_node.attrib["Originator"]
        self._maturity_level = fdl_node.attrib.get("MaturityLevel", default="Draft")
        self._locale = fdl_node.attrib.get("Locale", default="en-us")
        self._category = fdl_node.attrib.get("Category", default="none")
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{self._originator}/{self._category}/{self._identifier}/" f"v{self._feature_version.split('.')[0]}"
        )

        self._pb2_module, self._grpc_module = feature_definition_to_modules(fdl_node)
        self._servicer_cls = getattr(self._grpc_module, f"{self._identifier}Servicer")

        self._children_by_fully_qualified_identifier = {}

        self.defined_execution_errors = {}
        for err_node in xpath_sila(fdl_node, "sila:DefinedExecutionError"):
            err = DefinedExecutionErrorNode(err_node, self)
            self.defined_execution_errors[err._identifier] = err
            self._children_by_fully_qualified_identifier[err.fully_qualified_identifier] = err

        self._data_type_definitions = {}
        data_type_definition_nodes = list(xpath_sila(fdl_node, "sila:DataTypeDefinition"))
        failed_nodes = []
        while data_type_definition_nodes:
            num_nodes = len(data_type_definition_nodes)

            for dtype_node in data_type_definition_nodes:
                try:
                    dtype = DataTypeDefinition(dtype_node, self)
                    self._data_type_definitions[dtype._identifier] = dtype
                    self._children_by_fully_qualified_identifier[dtype.fully_qualified_identifier] = dtype
                except KeyError:
                    failed_nodes.append(dtype_node)

            if num_nodes == len(failed_nodes):  # pragma: no cover
                identifiers = [NamedNode(node)._identifier for node in failed_nodes]
                raise ValueError(
                    f"Feature definition contains cyclic dependencies between data type definitions {identifiers}"
                )

            data_type_definition_nodes = failed_nodes
            failed_nodes = []

        self._observable_properties = {}
        self._unobservable_properties = {}
        for prop_node in xpath_sila(fdl_node, "sila:Property"):
            prop = Property.from_fdl_node(prop_node, self)
            self._children_by_fully_qualified_identifier[prop.fully_qualified_identifier] = prop
            if isinstance(prop, ObservableProperty):
                self._observable_properties[prop._identifier] = prop
            elif isinstance(prop, UnobservableProperty):
                self._unobservable_properties[prop._identifier] = prop
            else:
                raise NotImplementedError  # pragma: no cover

        self._observable_commands = {}
        self._unobservable_commands = {}
        for command_node in xpath_sila(fdl_node, "sila:Command"):
            cmd = Command.from_fdl_node(command_node, self)
            self._children_by_fully_qualified_identifier[cmd.fully_qualified_identifier] = cmd
            for param in cmd.parameters:
                self._children_by_fully_qualified_identifier[param.fully_qualified_identifier] = param
            for resp in cmd.responses:
                self._children_by_fully_qualified_identifier[resp.fully_qualified_identifier] = resp
            if isinstance(cmd, ObservableCommand):
                self._observable_commands[cmd._identifier] = cmd
                if cmd.intermediate_responses is not None:
                    for int_resp in cmd.intermediate_responses:
                        self._children_by_fully_qualified_identifier[int_resp.fully_qualified_identifier] = int_resp
            elif isinstance(cmd, UnobservableCommand):
                self._unobservable_commands[cmd._identifier] = cmd
            else:
                raise NotImplementedError  # pragma: no cover

        self.metadata_definitions = {}
        for meta_node in xpath_sila(fdl_node, "sila:Metadata"):
            meta = Metadata(meta_node, self)
            self.metadata_definitions[meta._identifier] = meta
            self._children_by_fully_qualified_identifier[meta.fully_qualified_identifier] = meta

        self._binary_transfer_handler = None

    def __getitem__(self, item: str) -> Union[Property, Command, Metadata]:
        if item in self._unobservable_properties:
            return self._unobservable_properties[item]
        if item in self._observable_properties:
            return self._observable_properties[item]
        if item in self._unobservable_commands:
            return self._unobservable_commands[item]
        if item in self._observable_commands:
            return self._observable_commands[item]
        if item in self.metadata_definitions:
            return self.metadata_definitions[item]
        raise KeyError(f"Feature '{self._identifier}' has no child command, property or metadata '{item}'")



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Boolean as SilaBoolean


class Boolean(DataType):
    native_type = bool
    message_type: Type[SilaBoolean]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Boolean

    def to_native_type(self, message: SilaBoolean, toplevel_named_data_node: Optional[NamedDataNode] = None) -> bool:
        return message.value

    def to_message(self, value: bool, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaBoolean:
        if not isinstance(value, bool):
            raise TypeError("Expected a bool value")
        return self.message_type(value=value)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Type, Union

from google.protobuf.message import Message

from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.command import Command
from sila2.framework.data_types.binary import Binary
from sila2.framework.data_types.string import String
from sila2.framework.defined_execution_error_node import DefinedExecutionErrorNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.property.property import Property
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class Metadata(NamedDataNode):
    fully_qualified_identifier: FullyQualifiedIdentifier
    defined_execution_errors: List[DefinedExecutionErrorNode]
    parent_feature: Feature
    message_type: Type[Message]
    parameter_message_type: Type[Message]
    affected_calls_responses_message_type: Type[Message]

    def __init__(self, fdl_node, parent_feature: Feature):
        super().__init__(fdl_node, parent_feature, parent_feature._pb2_module)
        self.parent_feature = parent_feature
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_feature.fully_qualified_identifier}/Metadata/{self._identifier}"
        )
        self.defined_execution_errors = [
            parent_feature.defined_execution_errors[name]
            for name in xpath_sila(fdl_node, "sila:DefinedExecutionErrors/sila:Identifier/text()")
        ]

        self.message_type = getattr(self.parent_feature._pb2_module, f"Metadata_{self._identifier}")
        self.parameter_message_type = getattr(
            self.parent_feature._pb2_module,
            f"Get_FCPAffectedByMetadata_{self._identifier}_Parameters",
        )
        self.affected_calls_responses_message_type = getattr(
            self.parent_feature._pb2_module,
            f"Get_FCPAffectedByMetadata_{self._identifier}_Responses",
        )
        self.affected_calls_parameters_message_type = getattr(
            self.parent_feature._pb2_module,
            f"Get_FCPAffectedByMetadata_{self._identifier}_Responses",
        )
        self.__string_field = String(parent_feature._pb2_module.SiLAFramework__pb2)

    def get_parameter_message(self) -> Message:
        return self.parameter_message_type()

    def get_affected_calls_parameters_message(self) -> Message:
        return self.affected_calls_parameters_message_type()

    def to_affected_calls_message(
        self, affected_targets: List[Union[Feature, Property, Command, FullyQualifiedIdentifier]]
    ) -> Message:
        identifiers = []
        for target in affected_targets:
            if isinstance(target, str):
                identifiers.append(target)
            else:
                identifiers.append(target.fully_qualified_identifier)
        return self.affected_calls_responses_message_type(
            AffectedCalls=[self.__string_field.to_message(target) for target in identifiers]
        )

    def to_affected_calls_list(self, msg: Message) -> List[FullyQualifiedIdentifier]:
        return [
            FullyQualifiedIdentifier(self.__string_field.to_native_type(identifier)) for identifier in msg.AffectedCalls
        ]

    def to_message(self, value: Any, toplevel_named_data_node: Optional[NamedDataNode] = None) -> bytes:
        return self.message_type(
            **{self._identifier: self.data_type.to_message(value, toplevel_named_data_node=toplevel_named_data_node)}
        ).SerializeToString()

    def to_native_type(self, msg: bytes) -> Any:
        pb2_msg = self.message_type.FromString(msg)
        if isinstance(self.data_type, Binary) and getattr(pb2_msg, self._identifier).HasField("binaryTransferUUID"):
            raise ValueError("Cannot use Binary Transfer for SiLA Client Metadata")
        return self.data_type.to_native_type(getattr(pb2_msg, self._identifier))

    def to_grpc_header_key(self) -> str:
        return f"sila-{self.fully_qualified_identifier.lower().replace('/', '-')}-bin"



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Optional, Tuple, Type
from uuid import UUID

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.command.command_execution_uuid import CommandExecutionUUID
from sila2.framework.command.duration import Duration

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandConfirmation as SilaCommandConfirmation


class CommandConfirmation(MessageMappable):
    native_type = Tuple[UUID, Optional[timedelta]]
    message_type: Type[SilaCommandConfirmation]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.CommandConfirmation
        self.__duration_field = Duration(silaframework_pb2_module)
        self.__exec_id_field = CommandExecutionUUID(silaframework_pb2_module)

    def to_native_type(
        self, message: SilaCommandConfirmation, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> Tuple[UUID, Optional[timedelta]]:
        exec_id = self.__exec_id_field.to_native_type(message.commandExecutionUUID)
        if message.HasField("lifetimeOfExecution"):
            return exec_id, self.__duration_field.to_native_type(message.lifetimeOfExecution)
        else:
            return exec_id, None

    def to_message(
        self,
        execution_uuid: UUID,
        duration: Optional[timedelta] = None,
        toplevel_named_data_node: Optional[NamedDataNode] = None,
    ) -> SilaCommandConfirmation:
        exec_id = self.__exec_id_field.to_message(execution_uuid)
        if duration is None:
            return self.message_type(commandExecutionUUID=exec_id)
        return self.message_type(
            commandExecutionUUID=exec_id,
            lifetimeOfExecution=self.__duration_field.to_message(duration),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from os.path import join
from typing import TYPE_CHECKING, Type
from uuid import UUID

from sila2 import resource_dir
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.utils import run_protoc

pb2_module, grpc_module = run_protoc(join(resource_dir, "proto", "SiLABinaryTransfer.proto"))

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLABinaryTransfer_pb2, SiLABinaryTransfer_pb2_grpc
    from sila2.pb2_stubs.SiLAFramework_pb2 import Binary as SilaBinary

    pb2_module = SiLABinaryTransfer_pb2
    grpc_module = SiLABinaryTransfer_pb2_grpc


class BinaryTransferHandler(ABC):
    @abstractmethod
    def to_native_type(self, binary_uuid: UUID, toplevel_named_data_node: NamedDataNode) -> bytes:
        pass

    @abstractmethod
    def to_message(
        self, binary: bytes, message_type: Type[SilaBinary], toplevel_named_data_node: NamedDataNode
    ) -> SilaBinary:
        pass



from __future__ import annotations

from enum import Enum
from math import isclose
from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, Tuple

from sila2.framework.abc.constraint import Constraint
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.abc.data_type import DataType
    from sila2.framework.feature import Feature


class Unit(Constraint):
    label: str
    factor: float
    offset: float
    unit_components: Tuple[UnitComponent, ...]

    def __init__(
        self,
        label: str,
        factor: float,
        offset: float,
        unit_components: Optional[Iterable[UnitComponent]] = None,
    ):
        self.label = label
        self.factor = factor
        self.offset = offset
        self.unit_components = () if unit_components is None else tuple(unit_components)

    def validate(self, *args, **kwargs) -> bool:
        return True

    @classmethod
    def from_fdl_node(cls, fdl_node, parent_feature: Feature, base_type: DataType) -> Unit:
        label = xpath_sila(fdl_node, "sila:Label/text()")[0]
        factor = float(xpath_sila(fdl_node, "sila:Factor/text()")[0])
        offset = float(xpath_sila(fdl_node, "sila:Offset/text()")[0])

        unit_components = []
        for comp_node in xpath_sila(fdl_node, "sila:UnitComponent"):
            unit_str = xpath_sila(comp_node, "sila:SIUnit/text()")[0]
            exponent = getattr(SIUnit, xpath_sila(comp_node, "sila:SIUnit/text()")[0])
            unit_components.append(UnitComponent(getattr(SIUnit, unit_str), exponent))

        return cls(label, factor, offset, unit_components)

    def __repr__(self) -> str:
        factor_str = "" if isclose(self.factor, 1) else f" * {self.factor}"
        offset_str = "" if isclose(self.offset, 0) else f" + {self.offset}"
        return f"Unit({self.label!r}: '{' * '.join(str(u) for u in self.unit_components)}{factor_str}{offset_str}')"


class UnitComponent(NamedTuple):
    si_unit: SIUnit
    exponent: int

    def __str__(self) -> str:
        if self.exponent == 1:
            return self.si_unit.value
        return f"{self.si_unit.value}^{self.exponent}"


class SIUnit(Enum):
    Ampere = "A"
    Candela = "cd"
    Dimensionless = ""
    Kelvin = "K"
    Kilogram = "kg"
    Meter = "m"
    Mole = "mol"
    Second = "s"



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from sila2.framework.abc.sila_error import SilaError

if TYPE_CHECKING:
    from sila2.client.sila_client import SilaClient
    from sila2.pb2_stubs.SiLAFramework_pb2 import SiLAError
    from sila2.pb2_stubs.SiLAFramework_pb2 import UndefinedExecutionError as SilaUndefinedExecutionError


class UndefinedExecutionError(SilaError):
    message: str

    def __init__(self, message_or_exception: Union[str, Exception]):
        if isinstance(message_or_exception, Exception):
            self.message = f"{message_or_exception.__class__.__name__}: {message_or_exception}"
        else:
            self.message = message_or_exception
        super().__init__(self.message)

    def to_message(self) -> SiLAError:
        return self.pb2_module.SiLAError(
            undefinedExecutionError=self.pb2_module.UndefinedExecutionError(message=self.message)
        )

    @classmethod
    def from_message(cls, message: SilaUndefinedExecutionError, client: SilaClient) -> UndefinedExecutionError:
        return cls(message.message)



from typing import TYPE_CHECKING

from .authenticationservice_base import AuthenticationServiceBase
from .authenticationservice_errors import AuthenticationFailed, InvalidAccessToken
from .authenticationservice_feature import AuthenticationServiceFeature
from .authenticationservice_types import Login_Responses, Logout_Responses

__all__ = [
    "AuthenticationServiceBase",
    "AuthenticationServiceFeature",
    "Login_Responses",
    "Logout_Responses",
    "AuthenticationFailed",
    "InvalidAccessToken",
]

if TYPE_CHECKING:
    from .authenticationservice_client import AuthenticationServiceClient  # noqa: F401

    __all__.append("AuthenticationServiceClient")



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, List, TypeVar

from sila2.client.utils import call_rpc_function
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.metadata import Metadata

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature

T = TypeVar("T")


@dataclass
class ClientMetadataInstance(Generic[T]):
    metadata: Metadata
    value: T


class ClientMetadata(Generic[T]):
    _parent_feature: ClientFeature
    _wrapped_metadata: Metadata
    fully_qualified_identifier: FullyQualifiedIdentifier

    def __init__(self, parent_feature: ClientFeature, wrapped_metadata: Metadata):
        self._parent_feature = parent_feature
        self._wrapped_metadata = wrapped_metadata
        self.fully_qualified_identifier = self._wrapped_metadata.fully_qualified_identifier

    def __call__(self, value: T) -> ClientMetadataInstance[T]:
        return ClientMetadataInstance(self._wrapped_metadata, value)

    def get_affected_calls(self) -> List[FullyQualifiedIdentifier]:
        param_msg = self._wrapped_metadata.get_affected_calls_parameters_message()

        response_msg = call_rpc_function(
            getattr(self._parent_feature._grpc_stub, f"Get_FCPAffectedByMetadata_{self._wrapped_metadata._identifier}"),
            param_msg,
            metadata=None,
            client=self._parent_feature._parent_client,
            origin=None,
        )

        return self._wrapped_metadata.to_affected_calls_list(response_msg)



import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, List, Tuple
from uuid import UUID

from sila2.features.authenticationservice import (
    AuthenticationFailed,
    AuthenticationServiceBase,
    InvalidAccessToken,
    Login_Responses,
    Logout_Responses,
)
from sila2.framework import FullyQualifiedIdentifier
from sila2.server import SilaServer


@dataclass
class AccessToken:
    token: UUID
    features: Tuple[FullyQualifiedIdentifier]
    lifetime_period: timedelta
    last_usage: datetime

    @property
    def remaining_lifetime(self) -> timedelta:
        return self.lifetime_period - (datetime.now() - self.last_usage)


class AuthenticationServiceImpl(AuthenticationServiceBase):
    def __init__(self, parent_server: SilaServer):
        self.parent_server = parent_server
        self.active_tokens: Dict[UUID, AccessToken] = {}

    def Login(
        self,
        UserIdentification: str,
        Password: str,
        RequestedServer: str,
        RequestedFeatures: List[str],
        *,
        metadata: Dict[FullyQualifiedIdentifier, Any]
    ) -> Login_Responses:
        if UUID(RequestedServer) != self.parent_server.server_uuid:
            raise AuthenticationFailed

        # TODO: adapt to your needs
        if (UserIdentification, Password) != ("admin", "admin"):
            raise AuthenticationFailed

        server_feature_ids = {f.fully_qualified_identifier for f in self.parent_server.features.values()}
        for feature_id in RequestedFeatures:
            if FullyQualifiedIdentifier(feature_id) not in server_feature_ids:
                raise AuthenticationFailed

        token = AccessToken(uuid.uuid4(), tuple(server_feature_ids), timedelta(seconds=60 * 60), datetime.now())
        self.active_tokens[token.token] = token
        return Login_Responses(str(token.token), int(token.remaining_lifetime.total_seconds()))

    def Logout(self, AccessToken: str, *, metadata: Dict[FullyQualifiedIdentifier, Any]) -> Logout_Responses:
        try:
            token = self.active_tokens.pop(UUID(AccessToken))
        except (KeyError, ValueError):  # unknown token or no UUID
            raise InvalidAccessToken

        if token.remaining_lifetime.total_seconds() <= 0:
            self.active_tokens.pop(token.token)
            raise InvalidAccessToken



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from google.protobuf.message import Message

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier
from sila2.framework.utils import xpath_sila

if TYPE_CHECKING:
    from sila2.framework.feature import Feature


class DataTypeDefinition(NamedDataNode, DataType):
    parent_feature: Feature
    fully_qualified_identifier: FullyQualifiedIdentifier
    data_type: DataType

    def __init__(self, fdl_node, parent_feature: Feature):
        self.message_type = getattr(
            parent_feature._pb2_module,
            f"DataType_{xpath_sila(fdl_node, 'sila:Identifier/text()')[0]}",
        )
        super().__init__(fdl_node, parent_feature, self.message_type)
        self.parent_feature = parent_feature
        self.fully_qualified_identifier = FullyQualifiedIdentifier(
            f"{parent_feature.fully_qualified_identifier}/DataType/{self._identifier}"
        )
        self.native_type = self.data_type.native_type

    def to_native_type(self, message: Message, toplevel_named_data_node: Optional[NamedDataNode] = None) -> Any:
        return self.data_type.to_native_type(
            getattr(message, self._identifier), toplevel_named_data_node=toplevel_named_data_node
        )

    def to_message(self, *args: Any, **kwargs: Any) -> Message:
        return self.message_type(**{self._identifier: self.data_type.to_message(*args, **kwargs)})



from __future__ import annotations

from typing import NamedTuple


class Login_Responses(NamedTuple):

    AccessToken: str
    """
    The token to be used along with accessing a Command or Property on a SiLA Server.
    """

    TokenLifetime: int
    """
    The lifetime (in seconds) of the provided access token as the maximum validity period after the last SiLA Server
        request.
    """


class Logout_Responses(NamedTuple):

    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Iterable, Optional, TypeVar

from sila2.client.subscription_stream import SubscriptionStream
from sila2.client.utils import call_rpc_function
from sila2.framework.property.observable_property import ObservableProperty

if TYPE_CHECKING:
    from sila2.client.client_feature import ClientFeature
    from sila2.client.client_metadata import ClientMetadataInstance

T = TypeVar("T")


class ClientObservableProperty(Generic[T]):
    _parent_feature: ClientFeature
    _wrapped_property: ObservableProperty

    def __init__(self, parent_feature: ClientFeature, wrapped_property: ObservableProperty):
        self._parent_feature = parent_feature
        self._wrapped_property = wrapped_property

    def get(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> T:
        stream = self.subscribe(metadata=metadata)
        value = next(stream)
        stream.cancel()
        return value

    def subscribe(self, *, metadata: Optional[Iterable[ClientMetadataInstance]] = None) -> SubscriptionStream[T]:
        param_msg = self._wrapped_property.get_parameters_message()
        rpc_func = getattr(self._parent_feature._grpc_stub, f"Subscribe_{self._wrapped_property._identifier}")
        response_stream = call_rpc_function(
            rpc_func,
            param_msg,
            metadata=metadata,
            client=self._parent_feature._parent_client,
            origin=self._wrapped_property,
        )
        return SubscriptionStream(response_stream, self._wrapped_property.to_native_type)



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Dict, Iterable, Optional
from uuid import UUID, uuid4

from grpc import ServicerContext

from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module
from sila2.framework.binary_transfer.binary_upload_failed import BinaryUploadFailed
from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID
from sila2.framework.command.duration import Duration
from sila2.framework.utils import raise_as_rpc_error

if TYPE_CHECKING:
    from sila2.framework.binary_transfer.server_binary_transfer_handler import ServerBinaryTransferHandler
    from sila2.pb2_stubs import SiLABinaryTransfer_pb2


class BinaryUploadServicer(binary_transfer_grpc_module.BinaryUploadServicer):
    binaries_in_progress: Dict[UUID, Dict[int, Optional[bytes]]]
    _duration_field: Duration

    def __init__(self, parent_handler: ServerBinaryTransferHandler):
        self.parent_handler = parent_handler
        self.binaries_in_progress = {}
        self._duration_field = Duration(binary_transfer_pb2_module.SiLAFramework__pb2)

    def CreateBinary(self, request: SiLABinaryTransfer_pb2.CreateBinaryRequest, context: ServicerContext):
        try:
            bin_id = uuid4()
            chunk_count = request.chunkCount

            self.binaries_in_progress[bin_id] = {i: None for i in range(chunk_count)}

            return binary_transfer_pb2_module.CreateBinaryResponse(
                binaryTransferUUID=str(bin_id), lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1))
            )
        except Exception as ex:
            raise_as_rpc_error(BinaryUploadFailed(f"Upload of large binary failed: {ex}"), context)

    def UploadChunk(
        self, request_iterator: Iterable[SiLABinaryTransfer_pb2.UploadChunkRequest], context: ServicerContext
    ):
        try:
            for chunk_request in request_iterator:
                bin_id = UUID(chunk_request.binaryTransferUUID)

                if bin_id not in self.binaries_in_progress:
                    raise InvalidBinaryTransferUUID(f"Upload of large binary failed: invalid UUID {bin_id}")

                index = chunk_request.chunkIndex
                payload = chunk_request.payload

                self.binaries_in_progress[bin_id][index] = payload
                if all(isinstance(b, bytes) for b in self.binaries_in_progress[bin_id].values()):
                    self.parent_handler.known_binaries[bin_id] = b"".join(
                        self.binaries_in_progress[bin_id][i] for i in sorted(self.binaries_in_progress[bin_id].keys())
                    )
                    self.binaries_in_progress.pop(bin_id)

                yield binary_transfer_pb2_module.UploadChunkResponse(
                    binaryTransferUUID=str(bin_id),
                    chunkIndex=index,
                    lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1)),
                )
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryUploadFailed(f"Upload of large binary failed: {ex}"), context)

    def DeleteBinary(self, request: binary_transfer_pb2_module.DeleteBinaryRequest, context: ServicerContext):
        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Deletion of large binary failed: invalid UUID {bin_id}")

            self.parent_handler.known_binaries.pop(UUID(request.binaryTransferUUID))
            return binary_transfer_pb2_module.DeleteBinaryResponse()
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryUploadFailed(f"Deletion of large binary failed: {ex}"), context)



from __future__ import annotations

from threading import Thread
from typing import TYPE_CHECKING
from uuid import UUID

import grpc
from grpc import RpcError
from grpc._channel import _MultiThreadedRendezvous

from sila2.framework.command.execution_info import ExecutionInfo
from sila2.framework.fully_qualified_identifier import FullyQualifiedIdentifier

if TYPE_CHECKING:
    from sila2.client.client_observable_command_instance import ClientObservableCommandInstance


class ExecutionInfoSubscriptionThread(Thread):
    __stream: _MultiThreadedRendezvous
    __command_instance: ClientObservableCommandInstance
    __execution_info_field: ExecutionInfo

    def __init__(
        self,
        command_id: FullyQualifiedIdentifier,
        execution_uuid: UUID,
        stream: _MultiThreadedRendezvous,
        command_instance: ClientObservableCommandInstance,
    ):
        super().__init__(name=f"{self.__class__.__name__}-{command_id}-{execution_uuid}")
        self.__stream = stream
        self.__command_instance = command_instance
        self.__execution_info_field = ExecutionInfo(
            command_instance._client_command._wrapped_command.parent_feature._pb2_module.SiLAFramework__pb2
        )

    def run(self) -> None:
        try:
            for response_msg in self.__stream:
                execution_info = self.__execution_info_field.to_native_type(response_msg)
                self.__command_instance._update(execution_info)
        except RpcError as ex:
            if ex.code() == grpc.StatusCode.UNAVAILABLE:
                pass  # stream cancelled
            else:
                raise



class FeatureImplementationBase:
    def start(self) -> None:
        """This method is called by the SiLA server when it starts serving this feature implementation"""
        pass

    def stop(self) -> None:
        """This method is called by the SiLA server when it starts serving this feature implementation"""
        pass



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Iterable
from uuid import UUID

from grpc import ServicerContext

from sila2.framework.abc.binary_transfer_handler import grpc_module as binary_transfer_grpc_module
from sila2.framework.abc.binary_transfer_handler import pb2_module as binary_transfer_pb2_module
from sila2.framework.binary_transfer.binary_download_failed import BinaryDownloadFailed
from sila2.framework.binary_transfer.invalid_binary_transfer_uuid import InvalidBinaryTransferUUID
from sila2.framework.command.duration import Duration
from sila2.framework.utils import raise_as_rpc_error

if TYPE_CHECKING:
    from sila2.framework.binary_transfer.server_binary_transfer_handler import ServerBinaryTransferHandler


class BinaryDownloadServicer(binary_transfer_grpc_module.BinaryDownloadServicer):
    def __init__(self, parent_handler: ServerBinaryTransferHandler):
        self.parent_handler = parent_handler
        self._duration_field = Duration(binary_transfer_pb2_module.SiLAFramework__pb2)

    def GetBinaryInfo(self, request: binary_transfer_pb2_module.GetBinaryInfoRequest, context: ServicerContext):
        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

            return binary_transfer_pb2_module.GetBinaryInfoResponse(
                binarySize=len(self.parent_handler.known_binaries[bin_id]),
                lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1)),
            )
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryDownloadFailed(f"Download of large binary failed: {ex}"), context)

    def GetChunk(
        self, request_iterator: Iterable[binary_transfer_pb2_module.GetChunkRequest], context: ServicerContext
    ):
        try:
            for chunk_request in request_iterator:
                bin_id = UUID(chunk_request.binaryTransferUUID)

                if bin_id not in self.parent_handler.known_binaries:
                    raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

                offset = chunk_request.offset
                length = chunk_request.length

                yield binary_transfer_pb2_module.GetChunkResponse(
                    binaryTransferUUID=str(bin_id),
                    offset=offset,
                    payload=self.parent_handler.known_binaries[bin_id][offset : offset + length],
                    lifetimeOfBinary=self._duration_field.to_message(timedelta(minutes=1)),
                )
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryDownloadFailed(f"Download of large binary failed: {ex}"), context)

    def DeleteBinary(self, request: binary_transfer_pb2_module.DeleteBinaryRequest, context: ServicerContext):
        try:
            bin_id = UUID(request.binaryTransferUUID)

            if bin_id not in self.parent_handler.known_binaries:
                raise InvalidBinaryTransferUUID(f"Download of large binary failed: invalid UUID {bin_id}")

            self.parent_handler.known_binaries.pop(UUID(request.binaryTransferUUID))
            return binary_transfer_pb2_module.DeleteBinaryResponse()
        except InvalidBinaryTransferUUID as ex:
            raise_as_rpc_error(ex, context)
        except Exception as ex:
            raise_as_rpc_error(BinaryDownloadFailed(f"Deletion of large binary failed: {ex}"), context)



from __future__ import annotations

from datetime import time
from typing import TYPE_CHECKING, Optional, Type

from sila2.framework.abc.data_type import DataType
from sila2.framework.abc.named_data_node import NamedDataNode
from sila2.framework.data_types.timezone import Timezone

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import Time as SilaTime


class Time(DataType):
    native_type = time
    message_type: Type[SilaTime]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.Time
        self.__timezone_field = Timezone(silaframework_pb2_module)

    def to_message(self, t: time, toplevel_named_data_node: Optional[NamedDataNode] = None) -> SilaTime:
        if not isinstance(t, time):
            raise TypeError("Expected a time")
        if t.tzinfo is None:
            raise ValueError("No timezone provided")

        return self.message_type(
            hour=t.hour,
            minute=t.minute,
            second=t.second,
            timezone=self.__timezone_field.to_message(t.tzinfo),
        )

    def to_native_type(self, message: SilaTime, toplevel_named_data_node: Optional[NamedDataNode] = None) -> time:
        return time(
            hour=message.hour,
            minute=message.minute,
            second=message.second,
            tzinfo=self.__timezone_field.to_native_type(message.timezone),
        )

    @staticmethod
    def from_string(value: str) -> time:
        t = time.fromisoformat(value[:8])
        tz = Timezone.from_string(value[8:])
        return time(hour=t.hour, minute=t.minute, second=t.second, tzinfo=tz)



from sila2.framework.errors.framework_error import FrameworkError, FrameworkErrorType


class CommandExecutionNotFinished(FrameworkError):
    def __init__(self, message: str):
        super().__init__(FrameworkErrorType.COMMAND_EXECUTION_NOT_FINISHED, message)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type
from uuid import UUID

from sila2.framework.abc.message_mappable import MessageMappable
from sila2.framework.abc.named_data_node import NamedDataNode

if TYPE_CHECKING:
    from sila2.pb2_stubs import SiLAFramework_pb2
    from sila2.pb2_stubs.SiLAFramework_pb2 import CommandExecutionUUID as SilaCommandExecutionUUID


class CommandExecutionUUID(MessageMappable):
    native_type = UUID
    message_type: Type[SilaCommandExecutionUUID]

    def __init__(self, silaframework_pb2_module: SiLAFramework_pb2):
        self.message_type = silaframework_pb2_module.CommandExecutionUUID

    def to_native_type(
        self, message: SilaCommandExecutionUUID, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> UUID:
        return UUID(message.value)

    def to_message(
        self, value: UUID, toplevel_named_data_node: Optional[NamedDataNode] = None
    ) -> SilaCommandExecutionUUID:
        return self.message_type(value=str(value))



from sila2.client.client_feature import ClientFeature
from sila2.client.client_metadata import ClientMetadata, ClientMetadataInstance
from sila2.client.client_observable_command import ClientObservableCommand
from sila2.client.client_observable_command_instance import ClientObservableCommandInstance
from sila2.client.client_observable_property import ClientObservableProperty
from sila2.client.client_unobservable_command import ClientUnobservableCommand
from sila2.client.client_unobservable_property import ClientUnobservableProperty
from sila2.client.no_intermediate_responses import NoIntermediateResponses
from sila2.client.sila_client import SilaClient

__all__ = [
    "ClientFeature",
    "ClientMetadata",
    "ClientMetadataInstance",
    "ClientObservableCommand",
    "ClientObservableCommandInstance",
    "ClientObservableProperty",
    "ClientUnobservableCommand",
    "ClientUnobservableProperty",
    "NoIntermediateResponses",
    "SilaClient",
]
